# distutils: libraries = STD_LIBS
# distutils: include_dirs = LIB_DIR_GEOM
"""
A two-pass contour finding algorithm



"""

from __future__ import print_function

import numpy as np

cimport cython
cimport numpy as np
from libc.stdlib cimport free, malloc, realloc

from yt.geometry.oct_container cimport OctInfo, OctreeContainer
from yt.geometry.oct_visitors cimport Oct

from .amr_kdtools cimport Node
from .partitioned_grid cimport PartitionedGrid
from .volume_container cimport VolumeContainer, vc_index, vc_pos_index

import sys


cdef inline ContourID *contour_create(np.int64_t contour_id,
                               ContourID *prev = NULL):
    node = <ContourID *> malloc(sizeof(ContourID))
    #print("Creating contour with id", contour_id)
    node.contour_id = contour_id
    node.next = node.parent = NULL
    node.prev = prev
    node.count = 1
    if prev != NULL: prev.next = node
    return node

cdef inline void contour_delete(ContourID *node):
    if node.prev != NULL: node.prev.next = node.next
    if node.next != NULL: node.next.prev = node.prev
    free(node)

cdef inline ContourID *contour_find(ContourID *node):
    cdef ContourID *temp
    cdef ContourID *root
    root = node
    # First we find the root
    while root.parent != NULL and root.parent != root:
        root = root.parent
    if root == root.parent: root.parent = NULL
    # Now, we update everything along the tree.
    # So now everything along the line to the root has the parent set to the
    # root.
    while node.parent != NULL:
        temp = node.parent
        root.count += node.count
        node.count = 0
        node.parent = root
        node = temp
    return root

cdef inline void contour_union(ContourID *node1, ContourID *node2):
    if node1 == node2:
        return
    node1 = contour_find(node1)
    node2 = contour_find(node2)
    if node1 == node2:
        return
    cdef ContourID *pri
    cdef ContourID *sec
    if node1.count > node2.count:
        pri = node1
        sec = node2
    elif node2.count > node1.count:
        pri = node2
        sec = node1
    # might be a tie
    elif node1.contour_id < node2.contour_id:
        pri = node1
        sec = node2
    else:
        pri = node2
        sec = node1
    pri.count += sec.count
    sec.count = 0
    sec.parent = pri

cdef inline int candidate_contains(CandidateContour *first,
                            np.int64_t contour_id,
                            np.int64_t join_id = -1):
    while first != NULL:
        if first.contour_id == contour_id \
            and first.join_id == join_id: return 1
        first = first.next
    return 0

cdef inline CandidateContour *candidate_add(CandidateContour *first,
                                     np.int64_t contour_id,
                                     np.int64_t join_id = -1):
    cdef CandidateContour *node
    node = <CandidateContour *> malloc(sizeof(CandidateContour))
    node.contour_id = contour_id
    node.join_id = join_id
    node.next = first
    return node

cdef class ContourTree:
    # This class is essentially a Union-Find algorithm.  What we want to do is
    # to, given a connection between two objects, identify the unique ID for
    # those two objects.  So what we have is a collection of contours, and they
    # eventually all get joined and contain lots of individual IDs.  But it's
    # easy to find the *first* contour, i.e., the primary ID, for each of the
    # subsequent IDs.
    #
    # This means that we can connect id 202483 to id 2472, and if id 2472 is
    # connected to id 143, the connection will *actually* be from 202483 to
    # 143.  In this way we can speed up joining things and knowing their
    # "canonical" id.
    #
    # This is a multi-step process, since we first want to connect all of the
    # contours, then we end up wanting to coalesce them, and ultimately we join
    # them at the end.  The join produces a table that maps the initial to the
    # final, and we can go through and just update all of those.
    cdef ContourID *first
    cdef ContourID *last

    def clear(self):
        # Here, we wipe out ALL of our contours, but not the pointers to them
        cdef ContourID *cur
        cdef ContourID *next
        cur = self.first
        while cur != NULL:
            next = cur.next
            free(cur)
            cur = next
        self.first = self.last = NULL

    def __init__(self):
        self.first = self.last = NULL

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_contours(self, np.ndarray[np.int64_t, ndim=1] contour_ids):
        # This adds new contours, from the given contour IDs, to the tree.
        # Each one can be connected to a parent, as well as to next/prev in the
        # set of contours belonging to this tree.
        cdef int i, n
        n = contour_ids.shape[0]
        cdef ContourID *cur = self.last
        for i in range(n):
            #print(i, contour_ids[i])
            cur = contour_create(contour_ids[i], cur)
            if self.first == NULL: self.first = cur
        self.last = cur

    def add_contour(self, np.int64_t contour_id):
        self.last = contour_create(contour_id, self.last)

    def cull_candidates(self, np.ndarray[np.int64_t, ndim=3] candidates):
        # This function looks at each preliminary contour ID belonging to a
        # given collection of values, and then if need be it creates a new
        # contour for it.
        cdef int i, j, k, ni, nj, nk, nc
        cdef CandidateContour *first = NULL
        cdef CandidateContour *temp
        cdef np.int64_t cid
        nc = 0
        ni = candidates.shape[0]
        nj = candidates.shape[1]
        nk = candidates.shape[2]
        for i in range(ni):
            for j in range(nj):
                for k in range(nk):
                    cid = candidates[i,j,k]
                    if cid == -1: continue
                    if candidate_contains(first, cid) == 0:
                        nc += 1
                        first = candidate_add(first, cid)
        cdef np.ndarray[np.int64_t, ndim=1] contours
        contours = np.empty(nc, dtype="int64")
        i = 0
        # This removes all the temporary contours for this set of contours and
        # instead constructs a final list of them.
        while first != NULL:
            contours[i] = first.contour_id
            i += 1
            temp = first.next
            free(first)
            first = temp
        return contours

    def cull_joins(self, np.ndarray[np.int64_t, ndim=2] cjoins):
        # This coalesces contour IDs, so that we have only the final name
        # resolutions -- the .join_id from a candidate.  So many items will map
        # to a single join_id.
        cdef int i, ni, nc
        cdef CandidateContour *first = NULL
        cdef CandidateContour *temp
        cdef np.int64_t cid1, cid2
        nc = 0
        ni = cjoins.shape[0]
        for i in range(ni):
            cid1 = cjoins[i,0]
            cid2 = cjoins[i,1]
            if cid1 == -1: continue
            if cid2 == -1: continue
            if candidate_contains(first, cid1, cid2) == 0:
                nc += 1
                first = candidate_add(first, cid1, cid2)
        cdef np.ndarray[np.int64_t, ndim=2] contours
        contours = np.empty((nc,2), dtype="int64")
        i = 0
        while first != NULL:
            contours[i,0] = first.contour_id
            contours[i,1] = first.join_id
            i += 1
            temp = first.next
            free(first)
            first = temp
        return contours

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_joins(self, np.ndarray[np.int64_t, ndim=2] join_tree):
        cdef int i, n, ins
        cdef np.int64_t cid1, cid2
        # Okay, this requires lots of iteration, unfortunately
        cdef ContourID *cur
        cdef ContourID *c1
        cdef ContourID *c2
        n = join_tree.shape[0]
        #print("Counting")
        #print("Checking", self.count())
        for i in range(n):
            ins = 0
            cid1 = join_tree[i, 0]
            cid2 = join_tree[i, 1]
            c1 = c2 = NULL
            cur = self.first
            #print("Looking for ", cid1, cid2)
            while c1 == NULL or c2 == NULL:
                if cur.contour_id == cid1:
                    c1 = contour_find(cur)
                if cur.contour_id == cid2:
                    c2 = contour_find(cur)
                ins += 1
                cur = cur.next
                if cur == NULL: break
            if c1 == NULL or c2 == NULL:
                if c1 == NULL: print("  Couldn't find ", cid1)
                if c2 == NULL: print("  Couldn't find ", cid2)
                print("  Inspected ", ins)
                raise RuntimeError
            else:
                c1.count = c2.count = 0
                contour_union(c1, c2)

    def count(self):
        cdef int n = 0
        cdef ContourID *cur = self.first
        while cur != NULL:
            cur = cur.next
            n += 1
        return n

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def export(self):
        cdef int n = self.count()
        cdef ContourID *cur
        cdef ContourID *root
        cur = self.first
        cdef np.ndarray[np.int64_t, ndim=2] joins
        joins = np.empty((n, 2), dtype="int64")
        n = 0
        while cur != NULL:
            root = contour_find(cur)
            joins[n, 0] = cur.contour_id
            joins[n, 1] = root.contour_id
            cur = cur.next
            n += 1
        return joins

    def __dealloc__(self):
        self.clear()

cdef class TileContourTree:
    cdef np.float64_t min_val
    cdef np.float64_t max_val

    def __init__(self, np.float64_t min_val, np.float64_t max_val):
        self.min_val = min_val
        self.max_val = max_val

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def identify_contours(self, np.ndarray[np.float64_t, ndim=3] values,
                                np.ndarray[np.int64_t, ndim=3] contour_ids,
                                np.ndarray[np.uint8_t, ndim=3] mask,
                                np.int64_t start):
        # This just looks at neighbor values and tries to identify which zones
        # are touching by face within a given brick.
        cdef int i, j, k, ni, nj, nk, offset
        cdef int off_i, off_j, off_k, oi, ok, oj
        cdef ContourID *cur = NULL
        cdef ContourID *c1
        cdef ContourID *c2
        cdef np.float64_t v
        cdef np.int64_t nc
        ni = values.shape[0]
        nj = values.shape[1]
        nk = values.shape[2]
        nc = 0
        cdef ContourID **container = <ContourID**> malloc(
                sizeof(ContourID*)*ni*nj*nk)
        for i in range(ni*nj*nk): container[i] = NULL
        for i in range(ni):
            for j in range(nj):
                for k in range(nk):
                    v = values[i,j,k]
                    if mask[i,j,k] == 0: continue
                    if v < self.min_val or v > self.max_val: continue
                    nc += 1
                    c1 = contour_create(nc + start)
                    cur = container[i*nj*nk + j*nk + k] = c1
                    for oi in range(3):
                        off_i = oi - 1 + i
                        if not (0 <= off_i < ni): continue
                        for oj in range(3):
                            off_j = oj - 1 + j
                            if not (0 <= off_j < nj): continue
                            for ok in range(3):
                                if oi == oj == ok == 1: continue
                                off_k = ok - 1 + k
                                if not (0 <= off_k < nk): continue
                                if off_k > k and off_j > j and off_i > i:
                                    continue
                                offset = off_i*nj*nk + off_j*nk + off_k
                                c2 = container[offset]
                                if c2 == NULL: continue
                                c2 = contour_find(c2)
                                cur.count = c2.count = 0
                                contour_union(cur, c2)
                                cur = contour_find(cur)
        for i in range(ni):
            for j in range(nj):
                for k in range(nk):
                    c1 = container[i*nj*nk + j*nk + k]
                    if c1 == NULL: continue
                    c1 = contour_find(c1)
                    contour_ids[i,j,k] = c1.contour_id

        for i in range(ni*nj*nk):
            if container[i] != NULL: free(container[i])
        free(container)
        return nc

@cython.boundscheck(False)
@cython.wraparound(False)
def link_node_contours(Node trunk, contours, ContourTree tree,
        np.ndarray[np.int64_t, ndim=1] node_ids):
    cdef int n_nodes = node_ids.shape[0]
    cdef np.int64_t node_ind
    cdef VolumeContainer **vcs = <VolumeContainer **> malloc(
        sizeof(VolumeContainer*) * n_nodes)
    cdef int i
    cdef PartitionedGrid pg
    for i in range(n_nodes):
        pg = contours[node_ids[i]][2]
        vcs[i] = pg.container
    cdef np.ndarray[np.uint8_t] examined = np.zeros(n_nodes, "uint8")
    for _, cinfo in sorted(contours.items(), key = lambda a: -a[1][0]):
        _, node_ind, pg, _ = cinfo
        construct_boundary_relationships(trunk, tree, node_ind,
            examined, vcs, node_ids)
        examined[node_ind] = 1

cdef inline void get_spos(VolumeContainer *vc, int i, int j, int k,
                          int axis, np.float64_t *spos):
    spos[0] = vc.left_edge[0] + i * vc.dds[0]
    spos[1] = vc.left_edge[1] + j * vc.dds[1]
    spos[2] = vc.left_edge[2] + k * vc.dds[2]
    spos[axis] += 0.5 * vc.dds[axis]

cdef inline int spos_contained(VolumeContainer *vc, np.float64_t *spos):
    cdef int i
    for i in range(3):
        if spos[i] <= vc.left_edge[i] or spos[i] >= vc.right_edge[i]: return 0
    return 1

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef void construct_boundary_relationships(Node trunk, ContourTree tree,
                np.int64_t nid, np.ndarray[np.uint8_t, ndim=1] examined,
                VolumeContainer **vcs,
                np.ndarray[np.int64_t, ndim=1] node_ids):
    # We only look at the boundary and find the nodes next to it.
    # Contours is a dict, keyed by the node.id.
    cdef int i, j, off_i, off_j, oi, oj, ax, ax0, ax1, n1, n2
    cdef np.int64_t c1, c2
    cdef Node adj_node
    cdef VolumeContainer *vc1
    cdef VolumeContainer *vc0 = vcs[nid]
    cdef int s = (vc0.dims[1]*vc0.dims[0]
                + vc0.dims[0]*vc0.dims[2]
                + vc0.dims[1]*vc0.dims[2]) * 18
    # We allocate an array of fixed (maximum) size
    cdef np.ndarray[np.int64_t, ndim=2] joins = np.zeros((s, 2), dtype="int64")
    cdef int ti = 0, side, m1, m2, index
    cdef int pos[3]
    cdef int my_pos[3]
    cdef np.float64_t spos[3]

    for ax in range(3):
        ax0 = (ax + 1) % 3
        ax1 = (ax + 2) % 3
        n1 = vc0.dims[ax0]
        n2 = vc0.dims[ax1]
        for i in range(n1):
            for j in range(n2):
                for off_i in range(3):
                    oi = off_i - 1
                    if i == 0 and oi == -1: continue
                    if i == n1 - 1 and oi == 1: continue
                    for off_j in range(3):
                        oj = off_j - 1
                        if j == 0 and oj == -1: continue
                        if j == n2 - 1 and oj == 1: continue
                        pos[ax0] = i + oi
                        pos[ax1] = j + oj
                        my_pos[ax0] = i
                        my_pos[ax1] = j
                        for side in range(2):
                            # We go off each end of the block.
                            if side == 0:
                                pos[ax] = -1
                                my_pos[ax] = 0
                            else:
                                pos[ax] = vc0.dims[ax]
                                my_pos[ax] = vc0.dims[ax]-1
                            get_spos(vc0, pos[0], pos[1], pos[2], ax, spos)
                            adj_node = trunk._find_node(spos)
                            vc1 = vcs[adj_node.node_ind]
                            if spos_contained(vc1, spos):
                                index = vc_index(vc0, my_pos[0],
                                                 my_pos[1], my_pos[2])
                                m1 = vc0.mask[index]
                                c1 = (<np.int64_t*>vc0.data[0])[index]
                                index = vc_pos_index(vc1, spos)
                                m2 = vc1.mask[index]
                                c2 = (<np.int64_t*>vc1.data[0])[index]
                                if m1 == 1 and m2 == 1 and c1 > -1 and c2 > -1:
                                    if examined[adj_node.node_ind] == 0:
                                        joins[ti,0] = i64max(c1,c2)
                                        joins[ti,1] = i64min(c1,c2)
                                    else:
                                        joins[ti,0] = c1
                                        joins[ti,1] = c2
                                    ti += 1

    if ti == 0: return
    new_joins = tree.cull_joins(joins[:ti,:])
    tree.add_joins(new_joins)

@cython.boundscheck(False)
@cython.wraparound(False)
def update_joins(np.ndarray[np.int64_t, ndim=2] joins,
                 np.ndarray[np.int64_t, ndim=3] contour_ids,
                 np.ndarray[np.int64_t, ndim=1] final_joins):
    cdef int j, nj, nf
    cdef int ci, cj, ck
    nj = joins.shape[0]
    nf = final_joins.shape[0]
    for ci in range(contour_ids.shape[0]):
        for cj in range(contour_ids.shape[1]):
            for ck in range(contour_ids.shape[2]):
                if contour_ids[ci,cj,ck] == -1: continue
                for j in range(nj):
                    if contour_ids[ci,cj,ck] == joins[j,0]:
                        contour_ids[ci,cj,ck] = joins[j,1]
                        break
                for j in range(nf):
                    if contour_ids[ci,cj,ck] == final_joins[j]:
                        contour_ids[ci,cj,ck] = j + 1
                        break

cdef class FOFNode:
    cdef np.int64_t tag, count
    def __init__(self, np.int64_t tag):
        self.tag = tag
        self.count = 0

cdef class ParticleContourTree(ContourTree):
    cdef np.float64_t linking_length, linking_length2
    cdef np.float64_t DW[3]
    cdef np.float64_t DLE[3]
    cdef np.float64_t DRE[3]
    cdef bint periodicity[3]
    cdef int minimum_count

    def __init__(self, linking_length, periodicity = (True, True, True),
                 int minimum_count = 8):
        cdef int i
        self.linking_length = linking_length
        self.linking_length2 = linking_length * linking_length
        self.first = self.last = NULL
        for i in range(3):
            self.periodicity[i] = periodicity[i]
        self.minimum_count = minimum_count

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def identify_contours(self, OctreeContainer octree,
                                np.ndarray[np.int64_t, ndim=1] dom_ind,
                                np.ndarray[cython.floating, ndim=2] positions,
                                np.ndarray[np.int64_t, ndim=1] particle_ids,
                                int domain_id, int domain_offset):
        cdef np.ndarray[np.int64_t, ndim=1] pdoms, pcount, pind, doff
        cdef np.float64_t pos[3]
        cdef Oct *oct = NULL
        cdef Oct **neighbors = NULL
        cdef OctInfo oi
        cdef ContourID *c0
        cdef np.int64_t moff = octree.get_domain_offset(domain_id + domain_offset)
        cdef np.int64_t i, j, k, n, nneighbors = -1, pind0, offset
        cdef int counter = 0
        cdef int verbose = 0
        pcount = np.zeros_like(dom_ind)
        doff = np.zeros_like(dom_ind) - 1
        # First, we find the oct for each particle.
        pdoms = np.zeros(positions.shape[0], dtype="int64")
        pdoms -= -1
        # First we allocate our container
        cdef ContourID **container = <ContourID**> malloc(
            sizeof(ContourID*) * positions.shape[0])
        for i in range(3):
            self.DW[i] = (octree.DRE[i] - octree.DLE[i])
            self.DLE[i] = octree.DLE[i]
            self.DRE[i] = octree.DRE[i]
        for i in range(positions.shape[0]):
            counter += 1
            container[i] = NULL
            for j in range(3):
                pos[j] = positions[i, j]
            oct = octree.get(pos, NULL)
            if oct == NULL or (domain_id > 0 and oct.domain != domain_id):
                continue
            offset = oct.domain_ind - moff
            pcount[offset] += 1
            pdoms[i] = offset
        pind = np.argsort(pdoms)
        cdef np.int64_t *ipind = <np.int64_t*> pind.data
        cdef cython.floating *fpos = <cython.floating*> positions.data
        # pind is now the pointer into the position and particle_ids array.
        for i in range(positions.shape[0]):
            offset = pdoms[pind[i]]
            if doff[offset] < 0:
                doff[offset] = i
        del pdoms
        cdef int nsize = 27
        cdef np.int64_t *nind = <np.int64_t *> malloc(sizeof(np.int64_t)*nsize)
        counter = 0
        cdef np.int64_t frac = <np.int64_t> (doff.shape[0] / 20.0)
        if verbose == 1: print("Will be outputting every", frac, file=sys.stderr)
        for i in range(doff.shape[0]):
            if verbose == 1 and counter >= frac:
                counter = 0
                print("FOF-ing % 5.1f%% done" % ((100.0 * i)/doff.size), file=sys.stderr)
            counter += 1
            # Any particles found for this oct?
            if doff[i] < 0: continue
            offset = pind[doff[i]]
            # This can probably be replaced at some point with a faster lookup.
            for j in range(3):
                pos[j] = positions[offset, j]
            oct = octree.get(pos, &oi)
            if oct == NULL or (domain_id > 0 and oct.domain != domain_id):
                continue
            # Now we have our primary oct, so we will get its neighbors.
            neighbors = octree.neighbors(&oi, &nneighbors, oct,
                                self.periodicity)
            # Now we have all our neighbors.  And, we should be set for what
            # else we need to do.
            if nneighbors > nsize:
                nind = <np.int64_t *> realloc(
                    nind, sizeof(np.int64_t)*nneighbors)
                nsize = nneighbors
            for j in range(nneighbors):
                nind[j] = neighbors[j].domain_ind - moff
                for n in range(j):
                    if nind[j] == nind[n]:
                        nind[j] = -1
                    break
            # This is allocated by the neighbors function, so we deallocate it.
            free(neighbors)
            # We might know that all our internal particles are linked.
            # Otherwise, we look at each particle.
            for j in range(pcount[i]):
                # Note that this offset is the particle index
                pind0 = pind[doff[i] + j]
                # Look at each neighboring oct
                for k in range(nneighbors):
                    if nind[k] == -1: continue
                    offset = doff[nind[k]]
                    if offset < 0: continue
                    # NOTE: doff[i] will not monotonically increase.  So we
                    # need a unique ID for each container that we are
                    # accessing.
                    self.link_particles(container,
                                        fpos, ipind,
                                        pcount[nind[k]],
                                        offset, pind0,
                                        doff[i] + j)
        cdef np.ndarray[np.int64_t, ndim=1] contour_ids
        contour_ids = np.ones(positions.shape[0], dtype="int64")
        contour_ids *= -1
        # Perform one last contour_find on each.  Note that we no longer need
        # to look at any of the doff or internal offset stuff.
        for i in range(positions.shape[0]):
            if container[i] == NULL: continue
            container[i] = contour_find(container[i])
        for i in range(positions.shape[0]):
            if container[i] == NULL: continue
            c0 = container[i]
            if c0.count < self.minimum_count: continue
            contour_ids[i] = particle_ids[pind[c0.contour_id]]
        free(container)
        del pind
        return contour_ids

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void link_particles(self, ContourID **container,
                                   cython.floating *positions,
                                   np.int64_t *pind,
                                   np.int64_t pcount,
                                   np.int64_t noffset,
                                   np.int64_t pind0,
                                   np.int64_t poffset):
        # Now we look at each particle and evaluate it
        cdef np.float64_t pos0[3]
        cdef np.float64_t pos1[3]
        cdef np.float64_t edges[2][3]
        cdef int link
        cdef ContourID *c0
        cdef ContourID *c1
        cdef np.int64_t pind1
        cdef int i, j
        # We use pid here so that we strictly take new ones.
        # Note that pind0 will not monotonically increase, but
        c0 = container[pind0]
        if c0 == NULL:
            c0 = container[pind0] = contour_create(poffset, self.last)
            self.last = c0
            if self.first == NULL:
                self.first = c0
        c0 = container[pind0] = contour_find(c0)
        for i in range(3):
            # We make a very conservative guess here about the edges.
            pos0[i] = positions[pind0*3 + i]
            edges[0][i] = pos0[i] - self.linking_length*1.01
            edges[1][i] = pos0[i] + self.linking_length*1.01
            if edges[0][i] < self.DLE[i] or edges[0][i] > self.DRE[i]:
                # We skip this one, since we're close to the boundary
                edges[0][i] = -1e30
                edges[1][i] = 1e30
        # Lets set up some bounds for the particles.  Maybe we can get away
        # with reducing our number of calls to r2dist_early.
        for i in range(pcount):
            pind1 = pind[noffset + i]
            if pind1 == pind0: continue
            c1 = container[pind1]
            if c1 != NULL and c1.contour_id == c0.contour_id:
                # Already linked.
                continue
            for j in range(3):
                pos1[j] = positions[pind1*3 + j]
            link = r2dist_early(pos0, pos1, self.DW, self.periodicity,
                                self.linking_length2, edges)
            if link == 0: continue
            if c1 == NULL:
                c0.count += 1
                container[pind1] = c0
            elif c0.contour_id != c1.contour_id:
                contour_union(c0, c1)
                c0 = container[pind1] = container[pind0] = contour_find(c0)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int r2dist_early(np.float64_t ppos[3],
                             np.float64_t cpos[3],
                             np.float64_t DW[3],
                             bint periodicity[3],
                             np.float64_t max_r2,
                             np.float64_t edges[2][3]):
    cdef int i
    cdef np.float64_t r2, DR
    r2 = 0.0
    for i in range(3):
        if cpos[i] < edges[0][i]:
            return 0
        if cpos[i] > edges[1][i]:
            return 0
    for i in range(3):
        DR = (ppos[i] - cpos[i])
        if not periodicity[i]:
            pass
        elif (DR > DW[i]/2.0):
            DR -= DW[i]
        elif (DR < -DW[i]/2.0):
            DR += DW[i]
        r2 += DR * DR
        if r2 > max_r2: return 0
    return 1



import numpy as np

import yt
from yt.testing import requires_file
from yt.utilities.lib.bounding_volume_hierarchy import BVH, test_ray_trace
from yt.visualization.volume_rendering.api import Camera, Scene


def get_rays(camera):
    normal_vector = camera.unit_vectors[2].d
    W = np.array([8.0, 8.0])
    N = np.array([800, 800])
    dx = W / N

    x_points = np.linspace((-N[0] / 2 + 0.5) * dx[0], (N[0] / 2 - 0.5) * dx[0], N[0])
    y_points = np.linspace((-N[1] / 2 + 0.5) * dx[1], (N[1] / 2 - 0.5) * dx[0], N[1])

    X, Y = np.meshgrid(x_points, y_points)
    result = np.dot(camera.unit_vectors[0:2].T, [X.ravel(), Y.ravel()])
    vec_origins = camera.scene.arr(result.T, "unitary") + camera.position
    return np.array(vec_origins), np.array(normal_vector)


fn = "MOOSE_sample_data/out.e-s010"


@requires_file(fn)
def test_bounding_volume_hierarchy():
    ds = yt.load(fn)
    vertices = ds.index.meshes[0].connectivity_coords
    indices = ds.index.meshes[0].connectivity_indices - 1

    ad = ds.all_data()
    field_data = ad["connect1", "diffused"]

    bvh = BVH(vertices, indices, field_data)

    sc = Scene()
    cam = Camera(sc)
    cam.set_position(np.array([8.0, 8.0, 8.0]))
    cam.focus = np.array([0.0, 0.0, 0.0])
    origins, direction = get_rays(cam)

    image = np.empty(800 * 800, np.float64)
    test_ray_trace(image, origins, direction, bvh)
    image = image.reshape((800, 800))
    return image



import re

import numpy as np

from yt.frontends.sph.data_structures import ParticleDataset
from yt.utilities.chemical_formulas import ChemicalFormula
from yt.utilities.physical_ratios import _primordial_mass_fraction

from .field_plugin_registry import register_field_plugin

# See YTEP-0003 for details, but we want to ensure these fields are all
# populated:
#
#   * _mass
#   * _density
#   * _fraction
#   * _number_density
#


def _create_fraction_func(ftype, species):
    def _frac(field, data):
        return data[ftype, f"{species}_density"] / data[ftype, "density"]

    return _frac


def _mass_from_cell_volume_and_density(ftype, species):
    def _mass(field, data):
        return data[ftype, f"{species}_density"] * data["index", "cell_volume"]

    return _mass


def _mass_from_particle_mass_and_fraction(ftype, species):
    def _mass(field, data):
        return data[ftype, f"{species}_fraction"] * data[ftype, "particle_mass"]

    return _mass


def _create_number_density_func(ftype, species):
    formula = ChemicalFormula(species)

    def _number_density(field, data):
        weight = formula.weight  # This is in AMU
        weight *= data.ds.units.physical_constants.amu_cgs
        return data[ftype, f"{species}_density"] / weight

    return _number_density


def _create_density_func(ftype, species):
    def _density(field, data):
        return data[ftype, f"{species}_fraction"] * data[ftype, "density"]

    return _density


def add_species_field_by_density(registry, ftype, species):
    """
    This takes a field registry, a fluid type, and a species name and then
    adds the other fluids based on that.  This assumes that the field
    "SPECIES_density" already exists and refers to mass density.
    """
    unit_system = registry.ds.unit_system

    registry.add_field(
        (ftype, f"{species}_fraction"),
        sampling_type="local",
        function=_create_fraction_func(ftype, species),
        units="",
    )

    if isinstance(registry.ds, ParticleDataset):
        _create_mass_func = _mass_from_particle_mass_and_fraction
    else:
        _create_mass_func = _mass_from_cell_volume_and_density
    registry.add_field(
        (ftype, f"{species}_mass"),
        sampling_type="local",
        function=_create_mass_func(ftype, species),
        units=unit_system["mass"],
    )

    registry.add_field(
        (ftype, f"{species}_number_density"),
        sampling_type="local",
        function=_create_number_density_func(ftype, species),
        units=unit_system["number_density"],
    )

    return [
        (ftype, f"{species}_number_density"),
        (ftype, f"{species}_density"),
        (ftype, f"{species}_mass"),
    ]


def add_species_field_by_fraction(registry, ftype, species):
    """
    This takes a field registry, a fluid type, and a species name and then
    adds the other fluids based on that.  This assumes that the field
    "SPECIES_fraction" already exists and refers to mass fraction.
    """
    unit_system = registry.ds.unit_system

    registry.add_field(
        (ftype, f"{species}_density"),
        sampling_type="local",
        function=_create_density_func(ftype, species),
        units=unit_system["density"],
    )

    if isinstance(registry.ds, ParticleDataset):
        _create_mass_func = _mass_from_particle_mass_and_fraction
    else:
        _create_mass_func = _mass_from_cell_volume_and_density
    registry.add_field(
        (ftype, f"{species}_mass"),
        sampling_type="local",
        function=_create_mass_func(ftype, species),
        units=unit_system["mass"],
    )

    registry.add_field(
        (ftype, f"{species}_number_density"),
        sampling_type="local",
        function=_create_number_density_func(ftype, species),
        units=unit_system["number_density"],
    )

    return [
        (ftype, f"{species}_number_density"),
        (ftype, f"{species}_density"),
        (ftype, f"{species}_mass"),
    ]


def add_species_aliases(registry, ftype, alias_species, species):
    r"""
    This takes a field registry, a fluid type, and two species names.
    The first species name is one you wish to alias to an existing species
    name.  For instance you might alias all "H_p0" fields to "H\_" fields
    to indicate that "H\_" fields are really just neutral hydrogen fields.
    This function registers field aliases for the density, number_density,
    mass, and fraction fields between the two species given in the arguments.
    """
    registry.alias((ftype, f"{alias_species}_density"), (ftype, f"{species}_density"))
    registry.alias((ftype, f"{alias_species}_fraction"), (ftype, f"{species}_fraction"))
    registry.alias(
        (ftype, f"{alias_species}_number_density"),
        (ftype, f"{species}_number_density"),
    )
    registry.alias((ftype, f"{alias_species}_mass"), (ftype, f"{species}_mass"))


def add_deprecated_species_aliases(registry, ftype, alias_species, species):
    """
    Add the species aliases but with deprecation warnings.
    """

    for suffix in ["density", "fraction", "number_density", "mass"]:
        add_deprecated_species_alias(registry, ftype, alias_species, species, suffix)


def add_deprecated_species_alias(registry, ftype, alias_species, species, suffix):
    """
    Add a deprecated species alias field.
    """

    unit_system = registry.ds.unit_system
    if suffix == "fraction":
        my_units = ""
    else:
        my_units = unit_system[suffix]

    def _dep_field(field, data):
        return data[ftype, f"{species}_{suffix}"]

    registry.add_field(
        (ftype, f"{alias_species}_{suffix}"),
        sampling_type="local",
        function=_dep_field,
        units=my_units,
    )


def add_nuclei_density_fields(registry, ftype):
    unit_system = registry.ds.unit_system
    elements = _get_all_elements(registry.species_names)
    for element in elements:
        registry.add_field(
            (ftype, f"{element}_nuclei_density"),
            sampling_type="local",
            function=_nuclei_density,
            units=unit_system["number_density"],
        )

    # Here, we add default nuclei and number density fields for H and
    # He if they are not defined above, and if it was requested by
    # setting "default_species_fields"
    if registry.ds.default_species_fields is None:
        return

    dsf = registry.ds.default_species_fields
    # Right now, this only handles default fields for H and He
    for element in ["H", "He"]:
        # If these elements are already present in the dataset,
        # DO NOT set them
        if element in elements:
            continue
        # First add the default nuclei density fields
        registry.add_field(
            (ftype, f"{element}_nuclei_density"),
            sampling_type="local",
            function=_default_nuclei_density,
            units=unit_system["number_density"],
        )
        # Set up number density fields for hydrogen, either fully ionized or neutral.
        if element == "H":
            if dsf == "ionized":
                state = "p1"
            elif dsf == "neutral":
                state = "p0"
            else:
                raise NotImplementedError(
                    f"'default_species_fields' option '{dsf}' is not implemented!"
                )
            registry.alias(
                (ftype, f"H_{state}_number_density"), (ftype, "H_nuclei_density")
            )
        # Set up number density fields for helium, either fully ionized or neutral.
        if element == "He":
            if dsf == "ionized":
                state = "p2"
            elif dsf == "neutral":
                state = "p0"
            registry.alias(
                (ftype, f"He_{state}_number_density"), (ftype, "He_nuclei_density")
            )
    # If we're fully ionized, we need to setup the electron number density field
    if (ftype, "El_number_density") not in registry and dsf == "ionized":
        registry.add_field(
            (ftype, "El_number_density"),
            sampling_type="local",
            function=_default_nuclei_density,
            units=unit_system["number_density"],
        )


def _default_nuclei_density(field, data):
    ftype = field.name[0]
    element = field.name[1][: field.name[1].find("_")]
    amu_cgs = data.ds.units.physical_constants.amu_cgs
    if element == "El":
        # This is for determining the electron number density.
        # If we got here, this assumes full ionization!
        muinv = 1.0 * _primordial_mass_fraction["H"] / ChemicalFormula("H").weight
        muinv += 2.0 * _primordial_mass_fraction["He"] / ChemicalFormula("He").weight
    else:
        # This is for anything else besides electrons
        muinv = _primordial_mass_fraction[element] / ChemicalFormula(element).weight
    return data[ftype, "density"] * muinv / amu_cgs


def _nuclei_density(field, data):
    ftype = field.name[0]
    element = field.name[1][: field.name[1].find("_")]

    nuclei_mass_field = f"{element}_nuclei_mass_density"
    if (ftype, nuclei_mass_field) in data.ds.field_info:
        return (
            data[(ftype, nuclei_mass_field)]
            / ChemicalFormula(element).weight
            / data.ds.units.physical_constants.amu_cgs
        )
    metal_field = f"{element}_metallicity"
    if (ftype, metal_field) in data.ds.field_info:
        return (
            data[ftype, "density"]
            * data[(ftype, metal_field)]
            / ChemicalFormula(element).weight
            / data.ds.units.physical_constants.amu_cgs
        )

    field_data = np.zeros_like(
        data[ftype, f"{data.ds.field_info.species_names[0]}_number_density"]
    )
    for species in data.ds.field_info.species_names:
        nucleus = species
        if "_" in species:
            nucleus = species[: species.find("_")]
        # num is the number of nuclei contributed by this species.
        num = _get_element_multiple(nucleus, element)
        # Since this is a loop over all species existing in this dataset,
        # we will encounter species that contribute nothing, so we skip them.
        if num == 0:
            continue
        field_data += num * data[ftype, f"{species}_number_density"]
    return field_data


def _get_all_elements(species_list):
    elements = []
    for species in species_list:
        for item in re.findall("[A-Z][a-z]?|[0-9]+", species):
            if not item.isdigit() and item not in elements and item != "El":
                elements.append(item)
    return elements


def _get_element_multiple(compound, element):
    my_split = re.findall("[A-Z][a-z]?|[0-9]+", compound)
    if element not in my_split:
        return 0
    loc = my_split.index(element)
    if loc == len(my_split) - 1 or not my_split[loc + 1].isdigit():
        return 1
    return int(my_split[loc + 1])


@register_field_plugin
def setup_species_fields(registry, ftype="gas", slice_info=None):
    for species in registry.species_names:
        # These are all the species we should be looking for fractions or
        # densities of.
        if (ftype, f"{species}_density") in registry:
            func = add_species_field_by_density
        elif (ftype, f"{species}_fraction") in registry:
            func = add_species_field_by_fraction
        else:
            # Skip it
            continue
        func(registry, ftype, species)

        # Add aliases of X_p0_<field> to X_<field>.
        # These are deprecated and will be removed soon.
        if ChemicalFormula(species).charge == 0:
            alias_species = species.split("_")[0]
            if (ftype, f"{alias_species}_density") in registry:
                continue
            add_deprecated_species_aliases(registry, "gas", alias_species, species)

    add_nuclei_density_fields(registry, ftype)



from typing import List, Optional, Tuple, Union

import unyt as un
from numpy import ndarray

FieldDescT = Tuple[str, Tuple[str, List[str], Optional[str]]]
KnownFieldsT = Tuple[FieldDescT, ...]

ParticleType = str
FieldType = str
FieldName = str
FieldKey = Tuple[FieldType, FieldName]
ImplicitFieldKey = FieldName
AnyFieldKey = Union[FieldKey, ImplicitFieldKey]
DomainDimensions = Union[Tuple[int, ...], List[int], ndarray]

ParticleCoordinateTuple = Tuple[
    str,  # particle type
    Tuple[ndarray, ndarray, ndarray],  # xyz
    Union[float, ndarray],  # hsml
]

# Geometry specific types
AxisName = str
AxisOrder = Tuple[AxisName, AxisName, AxisName]

# types that can be converted to un.Unit
Unit = Union[un.Unit, str]

# types that can be converted to un.unyt_quantity
Quantity = Union[un.unyt_quantity, Tuple[float, Unit]]



import numpy as np
import pytest

from yt.utilities.cython_fortran_utils import FortranFile


def test_raise_error_when_file_does_not_exist():
    with pytest.raises(FileNotFoundError):
        FortranFile("/this/file/does/not/exist")


def test_read(tmp_path):
    dummy_file = tmp_path / "test.bin"
    # Write a Fortran-formatted file containing one record with 4 doubles
    dummy_file.write_bytes(
        b" \x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x08@\x00\x00\x00\x00\x00\x00\x10@ \x00\x00\x00"
    )
    with FortranFile(str(dummy_file)) as f:
        np.testing.assert_equal(
            f.read_vector("d"),
            [1.0, 2.0, 3.0, 4.0],
        )



import numpy as np
from numpy.testing import assert_array_equal, assert_equal

from yt.frontends.halo_catalog.data_structures import YTHaloCatalogDataset
from yt.frontends.ytdata.utilities import save_as_dataset
from yt.loaders import load as yt_load
from yt.testing import TempDirTest, requires_file, requires_module
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.answer_testing.framework import data_dir_load


def fake_halo_catalog(data):
    filename = "catalog.0.h5"

    ftypes = {field: "." for field in data}
    extra_attrs = {"data_type": "halo_catalog", "num_halos": data["particle_mass"].size}

    ds = {
        "cosmological_simulation": 1,
        "omega_lambda": 0.7,
        "omega_matter": 0.3,
        "hubble_constant": 0.7,
        "current_redshift": 0,
        "current_time": YTQuantity(1, "yr"),
        "domain_left_edge": YTArray(np.zeros(3), "cm"),
        "domain_right_edge": YTArray(np.ones(3), "cm"),
    }
    save_as_dataset(ds, filename, data, field_types=ftypes, extra_attrs=extra_attrs)
    return filename


class HaloCatalogTest(TempDirTest):
    @requires_module("h5py")
    def test_halo_catalog(self):
        rs = np.random.RandomState(3670474)
        n_halos = 100
        fields = ["particle_mass"] + [f"particle_position_{ax}" for ax in "xyz"]
        units = ["g"] + ["cm"] * 3
        data = {
            field: YTArray(rs.random_sample(n_halos), unit)
            for field, unit in zip(fields, units)
        }

        fn = fake_halo_catalog(data)
        ds = yt_load(fn)

        assert type(ds) is YTHaloCatalogDataset

        for field in fields:
            f1 = data[field].in_base()
            f1.sort()
            f2 = ds.r[("all", field)].in_base()
            f2.sort()
            assert_array_equal(f1, f2)

    @requires_module("h5py")
    def test_halo_catalog_boundary_particles(self):
        rs = np.random.RandomState(3670474)
        n_halos = 100
        fields = ["particle_mass"] + [f"particle_position_{ax}" for ax in "xyz"]
        units = ["g"] + ["cm"] * 3
        data = {
            field: YTArray(rs.random_sample(n_halos), unit)
            for field, unit in zip(fields, units)
        }

        data["particle_position_x"][0] = 1.0
        data["particle_position_x"][1] = 0.0
        data["particle_position_y"][2] = 1.0
        data["particle_position_y"][3] = 0.0
        data["particle_position_z"][4] = 1.0
        data["particle_position_z"][5] = 0.0

        fn = fake_halo_catalog(data)
        ds = yt_load(fn)

        assert type(ds) is YTHaloCatalogDataset

        for field in fields:
            f1 = data[field].in_base()
            f1.sort()
            f2 = ds.r[("all", field)].in_base()
            f2.sort()
            assert_array_equal(f1, f2)


t46 = "tiny_fof_halos/DD0046/DD0046.0.h5"


@requires_file(t46)
@requires_module("h5py")
def test_halo_quantities():
    ds = data_dir_load(t46)
    ad = ds.all_data()
    for i in range(ds.index.total_particles):
        hid = int(ad["halos", "particle_identifier"][i])
        halo = ds.halo("halos", hid)
        for field in ["mass", "position", "velocity"]:
            v1 = ad["halos", f"particle_{field}"][i]
            v2 = getattr(halo, field)
            assert_equal(v1, v2, err_msg=f"Halo {hid} {field} field mismatch.")


@requires_file(t46)
@requires_module("h5py")
def test_halo_particles():
    ds = data_dir_load(t46)
    i = ds.r["halos", "particle_mass"].argmax()
    hid = int(ds.r["halos", "particle_identifier"][i])
    halo = ds.halo("halos", hid)
    ids = halo["halos", "member_ids"]
    assert_equal(ids.size, 420)
    assert_equal(ids.min(), 19478.0)
    assert_equal(ids.max(), 31669.0)



import contextlib
import inspect
import shutil
import tempfile

from numpy.testing import assert_array_equal, assert_raises

import yt.units as u
from yt.config import ytcfg
from yt.loaders import load
from yt.testing import (
    assert_fname,
    fake_amr_ds,
    fake_hexahedral_ds,
    fake_random_ds,
    fake_tetrahedral_ds,
    requires_file,
    requires_module,
)
from yt.utilities.answer_testing.framework import (
    PlotWindowAttributeTest,
    data_dir_load,
    requires_ds,
)
from yt.utilities.exceptions import YTDataTypeUnsupported, YTPlotCallbackError
from yt.visualization.api import OffAxisSlicePlot, ProjectionPlot, SlicePlot
from yt.visualization.plot_container import accepts_all_fields

# These are a very simple set of tests that verify that each callback is or is
# not working.  They only check that it functions without an error; they do not
# check that it is providing correct results. Note that test_axis_manipulations
# is one exception, which is a full answer test.

# These are the callbacks still to test:
#
#  X velocity
#  X magnetic_field
#  X quiver
#  X contour
#  X grids
#  X streamlines
#    units
#  X line
#    cquiver
#    clumps
#  X arrow
#  X marker
#  X sphere
#    hop_circles
#    hop_particles
#    coord_axes
#  X text
#  X particles
#    title
#    flash_ray_data
#  X timestamp
#  X scale
#    material_boundary
#  X ray
#  X line_integral_convolution
#
#  X flip_horizontal, flip_vertical, swap_axes (all in test_axis_manipulations)

# cylindrical data for callback test
cyl_2d = "WDMerger_hdf5_chk_1000/WDMerger_hdf5_chk_1000.hdf5"
cyl_3d = "MHD_Cyl3d_hdf5_plt_cnt_0100/MHD_Cyl3d_hdf5_plt_cnt_0100.hdf5"


@contextlib.contextmanager
def _cleanup_fname():
    tmpdir = tempfile.mkdtemp()
    yield tmpdir
    shutil.rmtree(tmpdir)


def test_method_signature():
    ds = fake_amr_ds(
        fields=[("gas", "density"), ("gas", "velocity_x"), ("gas", "velocity_y")],
        units=["g/cm**3", "m/s", "m/s"],
    )
    p = SlicePlot(ds, "z", ("gas", "density"))
    sig = inspect.signature(p.annotate_velocity)
    # checking the first few arguments rather than the whole signature
    # we just want to validate that method wrapping works
    assert list(sig.parameters.keys())[:4] == [
        "factor",
        "scale",
        "scale_units",
        "normalize",
    ]


def test_init_signature_error_callback():
    ds = fake_amr_ds(
        fields=[("gas", "density"), ("gas", "velocity_x"), ("gas", "velocity_y")],
        units=["g/cm**3", "m/s", "m/s"],
    )
    p = SlicePlot(ds, "z", ("gas", "density"))
    # annotate_velocity accepts only one positional argument
    assert_raises(TypeError, p.annotate_velocity, 1, 2, 3)


def check_axis_manipulation(plot_obj, prefix):
    # convenience function for testing functionality of axis manipulation
    # callbacks. Can use in any of the other test functions.

    # test individual callbacks
    for cb in ("swap_axes", "flip_horizontal", "flip_vertical"):
        callback_handle = getattr(plot_obj, cb)
        callback_handle()  # toggles on for axis operation
        assert_fname(plot_obj.save(prefix)[0])
        callback_handle()  # toggle off

    # test all at once
    for cb in ("swap_axes", "flip_horizontal", "flip_vertical"):
        callback_handle = getattr(plot_obj, cb)
        callback_handle()

    assert_fname(plot_obj.save(prefix)[0])


def test_timestamp_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_timestamp()
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_timestamp()
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_timestamp()
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_timestamp(corner="lower_right", redshift=True, draw_inset_box=True)
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        p.annotate_timestamp(coord_system="axis")
        assert_fname(p.save(prefix)[0])


def test_timestamp_callback_code_units():
    # see https://github.com/yt-project/yt/issues/3869
    with _cleanup_fname() as prefix:
        ds = fake_random_ds(2, unit_system="code")
        p = SlicePlot(ds, "z", ("gas", "density"))
        p.annotate_timestamp()
        assert_fname(p.save(prefix)[0])


def test_scale_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = ProjectionPlot(ds, ax, ("gas", "density"), width=(0.5, 1.0))
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = ProjectionPlot(ds, ax, ("gas", "density"), width=(1.0, 1.5))
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_scale()
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_scale(corner="upper_right", coeff=10.0, unit="kpc")
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_scale(text_args={"size": 24})
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_scale(text_args={"font": 24})
        assert_raises(YTPlotCallbackError)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(YTDataTypeUnsupported, p.annotate_scale)
        assert_raises(YTDataTypeUnsupported, p.annotate_scale, coord_system="axis")


def test_line_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_line([0.1, 0.1, 0.1], [0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_line([0.1, 0.1, 0.1], [0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_line([0.1, 0.1, 0.1], [0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_line([0.1, 0.1], [0.5, 0.5], coord_system="axis", color="red")
        p.save(prefix)
        check_axis_manipulation(p, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(
            YTDataTypeUnsupported, p.annotate_line, [0.1, 0.1, 0.1], [0.5, 0.5, 0.5]
        )
        p.annotate_line([0.1, 0.1], [0.5, 0.5], coord_system="axis")
        assert_fname(p.save(prefix)[0])


def test_ray_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        ray = ds.ray((0.1, 0.2, 0.3), (0.6, 0.8, 0.5))
        oray = ds.ortho_ray(0, (0.3, 0.4))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_ray(oray)
        p.annotate_ray(ray)
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_ray(oray)
        p.annotate_ray(ray)
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_ray(oray)
        p.annotate_ray(ray)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_ray(oray)
        p.annotate_ray(ray, color="red")
        p.save(prefix)
        check_axis_manipulation(p, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        ray = ds.ray((0.1, 0.2, 0.3), (0.6, 0.8, 0.5))
        oray = ds.ortho_ray(0, (0.3, 0.4))
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(YTDataTypeUnsupported, p.annotate_ray, oray)
        assert_raises(YTDataTypeUnsupported, p.annotate_ray, ray)


def test_arrow_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_arrow([0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_arrow([0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_arrow([0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_arrow([0.5, 0.5], coord_system="axis", length=0.05)
        p.annotate_arrow(
            [[0.5, 0.6], [0.5, 0.6], [0.5, 0.6]], coord_system="data", length=0.05
        )
        p.annotate_arrow(
            [[0.5, 0.6, 0.8], [0.5, 0.6, 0.8], [0.5, 0.6, 0.8]],
            coord_system="data",
            length=0.05,
        )
        p.annotate_arrow(
            [[0.5, 0.6, 0.8], [0.5, 0.6, 0.8]], coord_system="axis", length=0.05
        )
        p.annotate_arrow(
            [[0.5, 0.6, 0.8], [0.5, 0.6, 0.8]], coord_system="figure", length=0.05
        )
        p.annotate_arrow(
            [[0.5, 0.6, 0.8], [0.5, 0.6, 0.8]], coord_system="plot", length=0.05
        )
        p.save(prefix)
        check_axis_manipulation(p, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(YTDataTypeUnsupported, p.annotate_arrow, [0.5, 0.5, 0.5])
        p.annotate_arrow([0.5, 0.5], coord_system="axis")
        assert_fname(p.save(prefix)[0])


def test_marker_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_marker([0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_marker([0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_marker([0.5, 0.5, 0.5])
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        coord = ds.arr([0.75, 0.75, 0.75], "unitary")
        coord.convert_to_units("kpc")
        p.annotate_marker(coord, coord_system="data")
        p.annotate_marker([0.5, 0.5], coord_system="axis", marker="*")
        p.annotate_marker([[0.5, 0.6], [0.5, 0.6], [0.5, 0.6]], coord_system="data")
        p.annotate_marker(
            [[0.5, 0.6, 0.8], [0.5, 0.6, 0.8], [0.5, 0.6, 0.8]], coord_system="data"
        )
        p.annotate_marker([[0.5, 0.6, 0.8], [0.5, 0.6, 0.8]], coord_system="axis")
        p.annotate_marker([[0.5, 0.6, 0.8], [0.5, 0.6, 0.8]], coord_system="figure")
        p.annotate_marker([[0.5, 0.6, 0.8], [0.5, 0.6, 0.8]], coord_system="plot")
        p.save(prefix)
        check_axis_manipulation(p, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(YTDataTypeUnsupported, p.annotate_marker, [0.5, 0.5, 0.5])
        p.annotate_marker([0.5, 0.5], coord_system="axis")
        assert_fname(p.save(prefix)[0])


def test_particles_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), particles=1)
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_particles((10, "Mpc"))
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_particles((10, "Mpc"))
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        ad = ds.all_data()
        p.annotate_particles(
            (10, "Mpc"),
            p_size=1.0,
            col="k",
            marker="o",
            stride=1,
            ptype="all",
            alpha=1.0,
            data_source=ad,
        )
        p.save(prefix)
        check_axis_manipulation(p, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(YTDataTypeUnsupported, p.annotate_particles, (10, "Mpc"))


def test_sphere_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_sphere([0.5, 0.5, 0.5], 0.1)
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_sphere([0.5, 0.5, 0.5], 0.1)
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_sphere([0.5, 0.5, 0.5], 0.1)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_sphere([0.5, 0.5], 0.1, coord_system="axis", text="blah")
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(
            YTDataTypeUnsupported,
            p.annotate_sphere,
            [0.5, 0.5, 0.5],
            0.1,
        )
        p.annotate_sphere([0.5, 0.5], 0.1, coord_system="axis", text="blah")
        assert_fname(p.save(prefix)[0])


def test_text_callback():
    with _cleanup_fname() as prefix:
        ax = "z"
        vector = [1.0, 1.0, 1.0]
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        p = ProjectionPlot(ds, ax, ("gas", "density"))
        p.annotate_text([0.5, 0.5, 0.5], "dinosaurs!")
        assert_fname(p.save(prefix)[0])
        p = SlicePlot(ds, ax, ("gas", "density"))
        p.annotate_text([0.5, 0.5, 0.5], "dinosaurs!")
        assert_fname(p.save(prefix)[0])
        p = OffAxisSlicePlot(ds, vector, ("gas", "density"))
        p.annotate_text([0.5, 0.5, 0.5], "dinosaurs!")
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_text(
            [0.5, 0.5], "dinosaurs!", coord_system="axis", text_args={"color": "red"}
        )
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        assert_raises(
            YTDataTypeUnsupported, p.annotate_text, [0.5, 0.5, 0.5], "dinosaurs!"
        )
        p.annotate_text(
            [0.5, 0.5], "dinosaurs!", coord_system="axis", text_args={"color": "red"}
        )
        assert_fname(p.save(prefix)[0])


@requires_module("h5py")
@requires_file(cyl_2d)
@requires_file(cyl_3d)
def test_velocity_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=("density", "velocity_x", "velocity_y", "velocity_z"),
            units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        )
        for ax in "xyz":
            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_velocity()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_velocity()
            assert_fname(p.save(prefix)[0])
        # Test for OffAxis Slice
        p = SlicePlot(ds, [1, 1, 0], ("gas", "density"), north_vector=[0, 0, 1])
        p.annotate_velocity(factor=40, normalize=True)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_velocity(factor=8, scale=0.5, scale_units="inches", normalize=True)
        assert_fname(p.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = fake_hexahedral_ds(fields=[f"velocity_{ax}" for ax in "xyz"])
        sl = SlicePlot(ds, 1, ("connect1", "test"))
        sl.annotate_velocity()
        assert_fname(sl.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "velocity_magnitude"))
        slc.annotate_velocity()
        assert_fname(slc.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = load(cyl_3d)
        for ax in ["r", "z", "theta"]:
            slc = SlicePlot(ds, ax, ("gas", "velocity_magnitude"))
            slc.annotate_velocity()
            assert_fname(slc.save(prefix)[0])
            slc = ProjectionPlot(ds, ax, ("gas", "velocity_magnitude"))
            slc.annotate_velocity()
            assert_fname(slc.save(prefix)[0])


def test_velocity_callback_spherical():
    ds = fake_amr_ds(
        fields=("density", "velocity_r", "velocity_theta", "velocity_phi"),
        units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        geometry="spherical",
    )

    with _cleanup_fname() as prefix:
        p = ProjectionPlot(ds, "phi", ("stream", "density"))
        p.annotate_velocity(factor=40, normalize=True)
        assert_fname(p.save(prefix)[0])

    with _cleanup_fname() as prefix:
        p = ProjectionPlot(ds, "r", ("stream", "density"))
        p.annotate_velocity(factor=40, normalize=True)
        assert_raises(NotImplementedError, p.save, prefix)


@requires_module("h5py")
@requires_file(cyl_2d)
@requires_file(cyl_3d)
def test_magnetic_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=(
                "density",
                "magnetic_field_x",
                "magnetic_field_y",
                "magnetic_field_z",
            ),
            units=(
                "g/cm**3",
                "G",
                "G",
                "G",
            ),
        )
        for ax in "xyz":
            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_magnetic_field()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_magnetic_field()
            assert_fname(p.save(prefix)[0])
        # Test for OffAxis Slice
        p = SlicePlot(ds, [1, 1, 0], ("gas", "density"), north_vector=[0, 0, 1])
        p.annotate_magnetic_field(factor=40, normalize=True)
        assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_magnetic_field(
            factor=8, scale=0.5, scale_units="inches", normalize=True
        )
        assert_fname(p.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "magnetic_field_strength"))
        slc.annotate_magnetic_field()
        assert_fname(slc.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = load(cyl_3d)
        for ax in ["r", "z", "theta"]:
            slc = SlicePlot(ds, ax, ("gas", "magnetic_field_strength"))
            slc.annotate_magnetic_field()
            assert_fname(slc.save(prefix)[0])
            slc = ProjectionPlot(ds, ax, ("gas", "magnetic_field_strength"))
            slc.annotate_magnetic_field()
            assert_fname(slc.save(prefix)[0])
        check_axis_manipulation(slc, prefix)  # only test the last axis

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=(
                "density",
                "magnetic_field_r",
                "magnetic_field_theta",
                "magnetic_field_phi",
            ),
            units=(
                "g/cm**3",
                "G",
                "G",
                "G",
            ),
            geometry="spherical",
        )
        p = ProjectionPlot(ds, "phi", ("gas", "density"))
        p.annotate_magnetic_field(
            factor=8, scale=0.5, scale_units="inches", normalize=True
        )
        assert_fname(p.save(prefix)[0])

        p = ProjectionPlot(ds, "theta", ("gas", "density"))
        p.annotate_magnetic_field(
            factor=8, scale=0.5, scale_units="inches", normalize=True
        )
        assert_fname(p.save(prefix)[0])

        p = ProjectionPlot(ds, "r", ("gas", "density"))
        p.annotate_magnetic_field(
            factor=8, scale=0.5, scale_units="inches", normalize=True
        )
        assert_raises(NotImplementedError, p.save, prefix)


@requires_module("h5py")
@requires_file(cyl_2d)
@requires_file(cyl_3d)
def test_quiver_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=("density", "velocity_x", "velocity_y", "velocity_z"),
            units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        )
        for ax in "xyz":
            p = ProjectionPlot(ds, ax, ("gas", "density"))
            p.annotate_quiver(("gas", "velocity_x"), ("gas", "velocity_y"))
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_quiver(("gas", "velocity_x"), ("gas", "velocity_y"))
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_quiver(("gas", "velocity_x"), ("gas", "velocity_y"))
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_quiver(
            ("gas", "velocity_x"),
            ("gas", "velocity_y"),
            factor=8,
            scale=0.5,
            scale_units="inches",
            normalize=True,
            bv_x=0.5 * u.cm / u.s,
            bv_y=0.5 * u.cm / u.s,
        )
        assert_fname(p.save(prefix)[0])
        check_axis_manipulation(p, prefix)

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "density"))
        slc.annotate_quiver(("gas", "velocity_r"), ("gas", "velocity_z"))
        assert_fname(slc.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = load(cyl_3d)
        slc = SlicePlot(ds, "r", ("gas", "velocity_magnitude"))
        slc.annotate_quiver(("gas", "velocity_theta"), ("gas", "velocity_z"))
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "z", ("gas", "velocity_magnitude"))
        slc.annotate_quiver(
            ("gas", "velocity_cartesian_x"), ("gas", "velocity_cartesian_y")
        )
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "theta", ("gas", "velocity_magnitude"))
        slc.annotate_quiver(("gas", "velocity_r"), ("gas", "velocity_z"))
        assert_fname(slc.save(prefix)[0])


def test_quiver_callback_spherical():
    ds = fake_amr_ds(
        fields=("density", "velocity_r", "velocity_theta", "velocity_phi"),
        units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        geometry="spherical",
    )

    with _cleanup_fname() as prefix:
        p = ProjectionPlot(ds, "phi", ("gas", "density"))
        p.annotate_quiver(
            ("gas", "velocity_cylindrical_radius"),
            ("gas", "velocity_cylindrical_z"),
            factor=8,
            scale=0.5,
            scale_units="inches",
            normalize=True,
        )
        assert_fname(p.save(prefix)[0])

    with _cleanup_fname() as prefix:
        p = ProjectionPlot(ds, "r", ("gas", "density"))
        p.annotate_quiver(
            ("gas", "velocity_theta"),
            ("gas", "velocity_phi"),
            factor=8,
            scale=0.5,
            scale_units="inches",
            normalize=True,
        )
        assert_fname(p.save(prefix)[0])


@requires_module("h5py")
@requires_file(cyl_2d)
def test_contour_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density", "temperature"), units=("g/cm**3", "K"))
        for ax in "xyz":
            p = ProjectionPlot(ds, ax, ("gas", "density"))
            p.annotate_contour(("gas", "temperature"))
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_contour(("gas", "temperature"))
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_contour(("gas", "temperature"))  # BREAKS WITH ndarray
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_contour(
            ("gas", "temperature"),
            levels=10,
            factor=8,
            take_log=False,
            clim=(0.4, 0.6),
            plot_args={"linewidths": 2.0},
            label=True,
            text_args={"fontsize": "x-large"},
        )
        p.save(prefix)

        p = SlicePlot(ds, "x", ("gas", "density"))
        s2 = ds.slice(0, 0.2)
        p.annotate_contour(
            ("gas", "temperature"),
            levels=10,
            factor=8,
            take_log=False,
            clim=(0.4, 0.6),
            plot_args={"linewidths": 2.0},
            label=True,
            text_args={"fontsize": "x-large"},
            data_source=s2,
        )
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "plasma_beta"))
        slc.annotate_contour(
            ("gas", "plasma_beta"),
            levels=2,
            factor=7,
            take_log=False,
            clim=(1.0e-1, 1.0e1),
            label=True,
            plot_args={"colors": ("c", "w"), "linewidths": 1},
            text_args={"fmt": "%1.1f"},
        )
        assert_fname(slc.save(prefix)[0])
        check_axis_manipulation(slc, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=("density", "temperature"),
            units=("g/cm**3", "K"),
            geometry="spherical",
        )
        p = SlicePlot(ds, "r", ("gas", "density"))
        kwargs = {
            "levels": 10,
            "factor": 8,
            "take_log": False,
            "clim": (0.4, 0.6),
            "plot_args": {"linewidths": 2.0},
            "label": True,
            "text_args": {"fontsize": "x-large"},
        }
        assert_raises(
            YTDataTypeUnsupported, p.annotate_contour, ("gas", "temperature"), **kwargs
        )


@requires_module("h5py")
@requires_file(cyl_2d)
def test_grids_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        for ax in "xyz":
            p = ProjectionPlot(ds, ax, ("gas", "density"))
            p.annotate_grids()
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_grids()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_grids()
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_grids(
            alpha=0.7,
            min_pix=10,
            min_pix_ids=30,
            draw_ids=True,
            id_loc="upper right",
            periodic=False,
            min_level=2,
            max_level=3,
            cmap="gist_stern",
        )
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "density"))
        slc.annotate_grids()
        assert_fname(slc.save(prefix)[0])
        check_axis_manipulation(slc, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = SlicePlot(ds, "r", ("gas", "density"))
        kwargs = {
            "alpha": 0.7,
            "min_pix": 10,
            "min_pix_ids": 30,
            "draw_ids": True,
            "id_loc": "upper right",
            "periodic": False,
            "min_level": 2,
            "max_level": 3,
            "cmap": "gist_stern",
        }
        assert_raises(YTDataTypeUnsupported, p.annotate_grids, **kwargs)


@requires_module("h5py")
@requires_file(cyl_2d)
def test_cell_edges_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
        for ax in "xyz":
            p = ProjectionPlot(ds, ax, ("gas", "density"))
            p.annotate_cell_edges()
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_cell_edges()
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_cell_edges()
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_cell_edges(alpha=0.7, line_width=0.9, color=(0.0, 1.0, 1.0))
        p.save(prefix)
        check_axis_manipulation(p, prefix)

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "density"))
        slc.annotate_cell_edges()
        assert_fname(slc.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), geometry="spherical")
        p = SlicePlot(ds, "r", ("gas", "density"))
        assert_raises(YTDataTypeUnsupported, p.annotate_cell_edges)


def test_mesh_lines_callback():
    with _cleanup_fname() as prefix:
        ds = fake_hexahedral_ds()
        for field in ds.field_list:
            sl = SlicePlot(ds, 1, field)
            sl.annotate_mesh_lines(color="black")
            assert_fname(sl.save(prefix)[0])

        ds = fake_tetrahedral_ds()
        for field in ds.field_list:
            sl = SlicePlot(ds, 1, field)
            sl.annotate_mesh_lines(color="black")
            assert_fname(sl.save(prefix)[0])
        check_axis_manipulation(sl, prefix)  # only test the final field


@requires_module("h5py")
@requires_file(cyl_2d)
@requires_file(cyl_3d)
def test_streamline_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=("density", "velocity_x", "velocity_y", "magvel"),
            units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        )

        for ax in "xyz":
            # Projection plot tests
            p = ProjectionPlot(ds, ax, ("gas", "density"))
            p.annotate_streamlines(("gas", "velocity_x"), ("gas", "velocity_y"))
            assert_fname(p.save(prefix)[0])

            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_streamlines(("gas", "velocity_x"), ("gas", "velocity_y"))
            assert_fname(p.save(prefix)[0])

            # Slice plot test
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_streamlines(("gas", "velocity_x"), ("gas", "velocity_y"))
            assert_fname(p.save(prefix)[0])

            # Additional features
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_streamlines(
                ("gas", "velocity_x"), ("gas", "velocity_y"), factor=32, density=4
            )
            assert_fname(p.save(prefix)[0])

            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_streamlines(
                ("gas", "velocity_x"),
                ("gas", "velocity_y"),
                field_color=("stream", "magvel"),
            )
            assert_fname(p.save(prefix)[0])
            check_axis_manipulation(p, prefix)

            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_streamlines(
                ("gas", "velocity_x"),
                ("gas", "velocity_y"),
                field_color=("stream", "magvel"),
                display_threshold=0.5,
                cmap="viridis",
                arrowstyle="->",
            )
            assert_fname(p.save(prefix)[0])

    # Axisymmetric dataset

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "velocity_magnitude"))
        slc.annotate_streamlines(("gas", "velocity_r"), ("gas", "velocity_z"))
        assert_fname(slc.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = load(cyl_3d)
        slc = SlicePlot(ds, "r", ("gas", "velocity_magnitude"))
        slc.annotate_streamlines(("gas", "velocity_theta"), ("gas", "velocity_z"))
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "z", ("gas", "velocity_magnitude"))
        slc.annotate_streamlines(
            ("gas", "velocity_cartesian_x"), ("gas", "velocity_cartesian_y")
        )
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "theta", ("gas", "velocity_magnitude"))
        slc.annotate_streamlines(("gas", "velocity_r"), ("gas", "velocity_z"))
        assert_fname(slc.save(prefix)[0])

    # Spherical dataset
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=("density", "velocity_r", "velocity_theta", "velocity_phi"),
            units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
            geometry="spherical",
        )
        slc = SlicePlot(ds, "phi", ("gas", "velocity_magnitude"))
        slc.annotate_streamlines(
            ("gas", "velocity_cylindrical_radius"), ("gas", "velocity_cylindrical_z")
        )
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "theta", ("gas", "velocity_magnitude"))
        slc.annotate_streamlines(
            ("gas", "velocity_conic_x"), ("gas", "velocity_conic_y")
        )
        assert_fname(slc.save(prefix)[0])


@requires_module("h5py")
@requires_file(cyl_2d)
@requires_file(cyl_3d)
def test_line_integral_convolution_callback():
    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=("density", "velocity_x", "velocity_y", "velocity_z"),
            units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
        )
        for ax in "xyz":
            p = ProjectionPlot(ds, ax, ("gas", "density"))
            p.annotate_line_integral_convolution(
                ("gas", "velocity_x"), ("gas", "velocity_y")
            )
            assert_fname(p.save(prefix)[0])
            p = ProjectionPlot(
                ds, ax, ("gas", "density"), weight_field=("gas", "density")
            )
            p.annotate_line_integral_convolution(
                ("gas", "velocity_x"), ("gas", "velocity_y")
            )
            assert_fname(p.save(prefix)[0])
            p = SlicePlot(ds, ax, ("gas", "density"))
            p.annotate_line_integral_convolution(
                ("gas", "velocity_x"), ("gas", "velocity_y")
            )
            assert_fname(p.save(prefix)[0])
        # Now we'll check a few additional minor things
        p = SlicePlot(ds, "x", ("gas", "density"))
        p.annotate_line_integral_convolution(
            ("gas", "velocity_x"),
            ("gas", "velocity_y"),
            kernellen=100.0,
            lim=(0.4, 0.7),
            cmap=ytcfg.get("yt", "default_colormap"),
            alpha=0.9,
            const_alpha=True,
        )
        p.save(prefix)

    with _cleanup_fname() as prefix:
        ds = load(cyl_2d)
        slc = SlicePlot(ds, "theta", ("gas", "magnetic_field_strength"))
        slc.annotate_line_integral_convolution(
            ("gas", "magnetic_field_r"), ("gas", "magnetic_field_z")
        )
        assert_fname(slc.save(prefix)[0])

    with _cleanup_fname() as prefix:
        ds = load(cyl_3d)
        slc = SlicePlot(ds, "r", ("gas", "magnetic_field_strength"))
        slc.annotate_line_integral_convolution(
            ("gas", "magnetic_field_theta"), ("gas", "magnetic_field_z")
        )
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "z", ("gas", "magnetic_field_strength"))
        slc.annotate_line_integral_convolution(
            ("gas", "magnetic_field_cartesian_x"), ("gas", "magnetic_field_cartesian_y")
        )
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "theta", ("gas", "magnetic_field_strength"))
        slc.annotate_line_integral_convolution(
            ("gas", "magnetic_field_r"), ("gas", "magnetic_field_z")
        )
        assert_fname(slc.save(prefix)[0])
        check_axis_manipulation(slc, prefix)

    with _cleanup_fname() as prefix:
        ds = fake_amr_ds(
            fields=("density", "velocity_r", "velocity_theta", "velocity_phi"),
            units=("g/cm**3", "cm/s", "cm/s", "cm/s"),
            geometry="spherical",
        )
        slc = SlicePlot(ds, "phi", ("gas", "velocity_magnitude"))
        slc.annotate_line_integral_convolution(
            ("gas", "velocity_cylindrical_radius"), ("gas", "velocity_cylindrical_z")
        )
        assert_fname(slc.save(prefix)[0])
        slc = SlicePlot(ds, "theta", ("gas", "velocity_magnitude"))
        slc.annotate_line_integral_convolution(
            ("gas", "velocity_conic_x"), ("gas", "velocity_conic_y")
        )
        assert_fname(slc.save(prefix)[0])
        check_axis_manipulation(slc, prefix)


def test_accepts_all_fields_decorator():
    fields = [
        ("gas", "density"),
        ("gas", "velocity_x"),
        ("gas", "pressure"),
        ("gas", "temperature"),
    ]
    units = ["g/cm**3", "cm/s", "dyn/cm**2", "K"]
    ds = fake_random_ds(16, fields=fields, units=units)
    plot = SlicePlot(ds, "z", fields=fields)

    # mocking a class method
    plot.fake_attr = {f: "not set" for f in fields}

    @accepts_all_fields
    def set_fake_field_attribute(self, field, value):
        self.fake_attr[field] = value
        return self

    # test on a single field
    plot = set_fake_field_attribute(plot, field=fields[0], value=1)
    assert_array_equal([plot.fake_attr[f] for f in fields], [1] + ["not set"] * 3)

    # test using "all" as a field
    plot = set_fake_field_attribute(plot, field="all", value=2)
    assert_array_equal(list(plot.fake_attr.values()), [2] * 4)


M7 = "DD0010/moving7_0010"


@requires_ds(M7)
def test_axis_manipulations():
    # tests flip_horizontal, flip_vertical and swap_axes in different combinations
    # on a SlicePlot with a velocity callback.
    plot_field = ("gas", "density")
    decimals = 12
    ds = data_dir_load(M7)

    def simple_velocity(test_obj, plot):
        # test_obj: the active PlotWindowAttributeTest
        # plot: the actual PlotWindow
        plot.annotate_velocity()

    def swap_axes(test_obj, plot):
        plot.swap_axes()

    def flip_horizontal(test_obj, plot):
        plot.flip_horizontal()

    def flip_vertical(test_obj, plot):
        plot.flip_vertical()

    callback_tests = (
        ("flip_horizontal", (simple_velocity, flip_horizontal)),
        ("flip_vertical", (simple_velocity, flip_vertical)),
        ("swap_axes", (simple_velocity, swap_axes)),
        ("flip_and_swap", (simple_velocity, flip_vertical, flip_horizontal, swap_axes)),
    )

    for n, r in callback_tests:
        test = PlotWindowAttributeTest(
            ds,
            plot_field,
            "x",
            attr_name=None,
            attr_args=None,
            decimals=decimals,
            callback_id=n,
            callback_runners=r,
        )
        test_axis_manipulations.__name__ = test.description
        yield test



import pytest

import yt
from yt.testing import fake_amr_ds


@pytest.mark.parametrize(
    "geometry",
    [
        "cartesian",
        "polar",
        "cylindrical",
        "spherical",
        "geographic",
        "internal_geographic",
        "spectral_cube",
    ],
)
def test_testable_geometries(geometry):
    # check that initializing a simple fake dataset works in any geometry
    ds = fake_amr_ds(fields=[("gas", "density")], units=["g/cm**3"], geometry=geometry)
    # make sure basic plotting works
    for axis in range(3):
        if "geographic" in geometry and axis == 2:
            pytest.skip(
                reason=(
                    "cannot test this case with vanilla yt (requires cartopy) "
                    "see https://github.com/yt-project/yt/issues/4182"
                )
            )
        yt.SlicePlot(ds, axis, ("gas", "density"), buff_size=(8, 8))
